#include <kytea/kytea-model.h>
#include "liblinear/linear.h"
#include <cstdlib>
#include <cmath>
#include <iostream>

using namespace kytea;
using namespace std;

#define SIG_CUTOFF 1E-6

pair<int,double> KyteaModel::runClassifier(const vector<unsigned> & feat) {
    unsigned i, j, idx, n = getNumFeatures(), numClass = labels_.size();
    // add weights
    vector< double > decs(numW_);
    for(i = 0; i < decs.size(); i++)
        decs[i] = 0.0;
    for(i = 0; i < feat.size(); i++) {
        idx = feat[i]-1;
        if(idx < n)
            for(j = 0; j < numW_; j++)
                decs[j] += getWeight(idx,j);
    }
    if(bias_) {
        idx = getBiasId()-1;
        for(j = 0; j < numW_; j++)
            decs[j] += getWeight(idx,j);
    }
    // find the return value
    pair<int,double> ret;
    if(numW_==1) {
        ret.first = (decs[0]>0?labels_[0]:labels_[1]);
        ret.second = abs(decs[0]);
    } else {
        unsigned maxId = 0;
        double diff = 0;
        for(i=1;i<numW_;i++) {
            if(decs[i] >= decs[maxId]) {
                diff = decs[i] - decs[maxId];
                maxId = i;
            }
        }
        ret.first = labels_[maxId];
        ret.second = diff;
    }
    return ret;
}

// allocate a LL feature node
feature_node * allocateFeatures(const vector<unsigned> & feats, int biasId, double biasVal) {
    feature_node * nodes = (feature_node*)malloc((feats.size()+(biasVal>=0?2:1))*sizeof(feature_node));
    unsigned i;
    for(i = 0; i < feats.size(); i++) {
        nodes[i].index=feats[i];
        nodes[i].value = 1;
    }
    nodes[i].index = biasId;
    nodes[i++].value = biasVal;
    nodes[i].index = -1;
    return nodes;
}
// train the model
void KyteaModel::trainModel(const vector< vector<unsigned> > & xs, vector<int> & ys, bool bias, int solver, double epsilon) {
    if(weights_.size()>0)
        weights_.clear();
    // build the liblinear model
    struct problem   prob;
    struct parameter param;
    prob.l = xs.size();
    prob.y = &ys.front();

    // allocate the feature space
    feature_node** myXs = (feature_node**)malloc(sizeof(feature_node*)*xs.size());
    int biasId = getBiasId();
    for(int i = 0; i < prob.l; i++)
        myXs[i] = allocateFeatures(xs[i], biasId, (bias?1:0));
    prob.x = myXs;

    prob.bias = (bias?1:-1);
    prob.n = names_.size()+(bias?1:0);

    param.solver_type = solver;
    param.C = 1; // L2R_L2LOSS_SVC_DUAL;
    param.eps = epsilon;
    param.nr_weight = 0;
    param.weight_label = NULL;
    param.weight = NULL;
    if(param.eps == HUGE_VAL) {
    	if(param.solver_type == L2R_LR || param.solver_type == L2R_L2LOSS_SVC)
    		param.eps = 0.01;
    	else if(param.solver_type == L2R_L2LOSS_SVC_DUAL || param.solver_type == L2R_L1LOSS_SVC_DUAL || param.solver_type == MCSVM_CS)
    		param.eps = 0.1;
    	else if(param.solver_type == L1R_L2LOSS_SVC || param.solver_type == L1R_LR)
    		param.eps = 0.01;
    }
    model* mod_ = train(&prob, &param);

    // free the problem
    for(int i = 0; i < prob.l; i++)
        free(myXs[i]);
    free(myXs);

    int i, j;

    // create the labels
    labels_.resize(mod_->nr_class);
    for(int i = 0; i < mod_->nr_class; i++)
        labels_[i] = mod_->label[i];

    // trim values
    FeatVec oldNames = names_;
    names_.clear();
    ids_.clear();
    KyteaString empty;
    mapFeat(empty);
    weights_.clear();
    numW_ = (labels_.size()==2 && solver_ != MCSVM_CS?1:labels_.size());
    for(i=0; i<oldNames.size()-1; i++) {
        double myMax = 0.0;
    	for(j=0; j<numW_; j++) 
            myMax = max(abs(mod_->w[i*numW_+j]),myMax);
        if(myMax>SIG_CUTOFF) {
            unsigned idx = mapFeat(oldNames[i+1]);
            for(j = 0; j < numW_; j++)
                weights_.push_back(mod_->w[i*numW_+j]);
        }
    }
    if(bias_)
        for(j = 0; j < numW_; j++)
            weights_.push_back(mod_->w[i*numW_+j]);

    destroy_model(mod_);

}

void KyteaModel::setNumClasses(unsigned v) {
    labels_.resize(v);
    numW_ = (v==2 && solver_ != MCSVM_CS?1:v);
}
