/*
* Copyright 2009, KyTea Development Team
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* 
*     http://www.apache.org/licenses/LICENSE-2.0
* 
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#include <set>
#include <cmath>
#include <kytea/kytea.h>
#include <kytea/corpus-io.h>
#include <kytea/model-io.h>
#include <kytea/dictionary.h>

using namespace kytea;
using namespace std;

typedef Dictionary::WordMap WordMap;
void addPron(WordMap& allWords, const KyteaString & word, const KyteaString & pron, bool dict) {
    WordMap::iterator it = allWords.find(word);
    if(it == allWords.end()) {
        DictionaryEntry * ent = new DictionaryEntry(word);
        if(pron.length() != 0)
            ent->prons.push_back(pron);
        ent->inDict = dict;
        allWords.insert(WordMap::value_type(word,ent));
    }
    else if(pron.length() != 0) {
        vector<KyteaString> & prons = it->second->prons;
        if(prons.size() == 0)
            prons.push_back(pron);
        else if(!dict) {
            unsigned i;
            for(i = 0; i < prons.size() && prons[i] != pron; i++);
            if(i == prons.size())
                prons.push_back(pron);
        }
    }
}
void Kytea::buildVocabulary() {

    WordMap allWords;

    if(config_->getDebug() > 0)
        cerr << "Scanning dictionaries and corpora for vocabulary" << endl;
    // scan the dictionaries
    const vector<string> & dict = config_->getDictionaryFiles();
    KyteaString slashString = util_->mapString("/"), word, pron;
    for(vector<string>::const_iterator it = dict.begin(); it != dict.end(); it++) {
        if(config_->getDebug() > 0)
            cerr << "Reading dictionary from " << *it << " ";
        FullCorpusIO io(util_, it->c_str(),false);
        ifstream dis(it->c_str());
        KyteaSentence* next;
        while((next = io.readSentence())) {
            if(next->words.size() != 1)
                throw runtime_error("Badly formatted dictionary entry (too many or too few words)");
            word = next->words[0].surf;
            pron = next->words[0].pron;
            addPron(allWords, word, pron, true);
            delete next;
        }
        if(config_->getDebug() > 0)
            cerr << "done!" <<endl;
    }
    
    vector<string> corpora = config_->getCorpusFiles();
    vector<CorpusIO::Format> corpForm = config_->getCorpusFormats();
    for(unsigned i = 0; i < corpora.size(); i++) {
        if(config_->getDebug() > 0)
            cerr << "Reading corpus from " << corpora[i] << " ";
        CorpusIO * io = CorpusIO::createIO(corpora[i].c_str(), corpForm[i], false, util_);
        KyteaSentence* next;
        while((next = io->readSentence())) {
            for(unsigned i = 0; i < next->words.size(); i++)
                if(next->words[i].isCertain)
                    addPron(allWords, next->words[i].surf, next->words[i].pron, false);
            sentences_.push_back(next);
        }
        if(config_->getDebug() > 0)
            cerr << " done!" << endl;
        delete io;
    }

    if(config_->getDebug() > 0)
        cerr << "Building dictionary index ";
    dict_->buildIndex(allWords);
    if(config_->getDebug() > 0)
        cerr << "done!" << endl;

}

// TODO skip calculating unmarked features?
unsigned Kytea::wsDictionaryFeatures(const KyteaString & chars, SentenceFeatures & features, const vector<unsigned> & featNums) {
    vector<DictionaryEntry*> & entries = dict_->getEntries();
    vector<DictionaryState*> & states = dict_->getStates();
    DictionaryEntry* myEntry;
    unsigned currState = 0, nextState;
    const unsigned len = features.size(), max=config_->getDictionaryN();
    char* on = (char*)malloc(len*3*max);
    unsigned ret;
    memset(on,0,len*3*max);
    for(unsigned i = 0; i < len+1; i++) {
        KyteaChar c = chars[i];
        while((nextState = states[currState]->step(c)) == 0 && currState != 0)
            currState = states[currState]->failure;
        currState = nextState;
        vector<unsigned> & output = states[currState]->output;
        for(unsigned j = 0; j < output.size(); j++) {
            myEntry = entries[output[j]];
            if(!myEntry->inDict)
                continue;
            const unsigned wlen = myEntry->word.length();
            const unsigned lablen = min(wlen,max)-1;
            // left value (position i-wlen, type
            if(i >= wlen)
                on[(i-wlen)*3*max +/*0*max*/+ lablen] = 1;
            // right value
            if(i != len)
                on[    i*3*max    +  2*max  + lablen] = 1;
            // middle values
            for(unsigned k = i-wlen+1; k < i; k++)
                on[    k*3*max    +  1*max  + lablen] = 1;
        }
    }
    for(unsigned i = 0; i < len; i++) {
        char* myOn = on+i*3*max;
        for(unsigned j = 0; j < 3*max; j++)
            if(myOn[j]) {
                features[i].push_back(featNums[j]);
                ret++;
            }
    }
    free(on);
    return ret;
}
unsigned Kytea::wsNgramFeatures(const KyteaString & chars, SentenceFeatures & features, const vector<KyteaString> & prefixes, int n) {
    int w = (int)prefixes.size()/2, rightBound;
    unsigned ret = 0;
    for(int i = 0; i < (int)features.size(); i++) {
        rightBound=min(i+w+1,(int)chars.length());
        for(int j = i-w+1; j < rightBound; j++) {
            if(j < 0) continue;
            KyteaString str = prefixes[j-i+w-1];
            for(int k = 0; k < n && j+k<rightBound; k++) {
                str = str+chars[j+k];
                features[i].push_back(wsModel_->mapFeat(str));
                ret++;
            }
        }
    }
    return ret;
}

void Kytea::prepareDictPrefixes(vector<unsigned> & dictFeats) {
    const char cs[3] = { 'L', 'I', 'R' };
    for(unsigned i = 0; i < 3; i++) {
        for(unsigned j = 0; j < (unsigned)config_->getDictionaryN(); j++) {
            ostringstream buff;
            buff << cs[i] << (j+1);
            dictFeats.push_back(wsModel_->mapFeat(util_->mapString(buff.str())));
        }
    }
}
void Kytea::prepareNgramPrefixes(vector<KyteaString> & charPrefixes, vector<KyteaString> & typePrefixes) {
    // create n-gram feature prefixes
    for(int i = 1; i <= 2*(int)config_->getCharWindow(); i++) {
        ostringstream buff;
        buff << "X" << i-(int)config_->getCharWindow();
        charPrefixes.push_back(util_->mapString(buff.str()));
    }
    for(int i = 1; i <= 2*(int)config_->getTypeWindow(); i++) {
        ostringstream buff;
        buff << "T" << i-(int)config_->getTypeWindow();
        typePrefixes.push_back(util_->mapString(buff.str()));
    }
}
void Kytea::trainWS() {
    if(wsModel_)
        delete wsModel_;
    wsModel_ = new KyteaModel();
    if(config_->getDebug() > 0)
        cerr << "Creating word segmentation features ";
    // create word prefixes
    vector<unsigned> dictFeats;
    vector<KyteaString> charPrefixes, typePrefixes;
    bool hasDictionary = (config_->getDictionaryFiles().size() > 0 && dict_->getStates().size() > 0);
    prepareDictPrefixes(dictFeats);
    prepareNgramPrefixes(charPrefixes,typePrefixes);
    // make the sentence features one by one
    unsigned scount = 0;
    vector< vector<unsigned> > xs;
    vector<int> ys;
    for(Sentences::const_iterator it = sentences_.begin(); it != sentences_.end(); it++) {
        if(++scount % 1000 == 0)
            cerr << ".";
        KyteaSentence * sent = *it;
        SentenceFeatures feats(sent->wsConfs.size());
        unsigned fts = 0;
        if(hasDictionary)
            fts += wsDictionaryFeatures(sent->chars, feats, dictFeats);
        fts += wsNgramFeatures(sent->chars, feats, charPrefixes, config_->getCharN());
        string str = util_->getTypeString(sent->chars);
        fts += wsNgramFeatures(util_->mapString(str), feats, typePrefixes, config_->getTypeN());
        for(unsigned i = 0; i < feats.size(); i++) {
            if(abs(sent->wsConfs[i]) > config_->getConfidence()) {
                xs.push_back(feats[i]);
                ys.push_back(sent->wsConfs[i]>1?1:-1);
            }
        }
    }
    if(config_->getDebug() > 0)
        cerr << " done!" << endl << "Building classifier ";
    
    wsModel_->trainModel(xs,ys,config_->hasBias(),config_->getSolverType(),config_->getEpsilon());

    if(config_->getDebug() > 0)
        cerr << " done!" << endl;

}


unsigned Kytea::peCharFeatures(const KyteaString & chars, vector<unsigned> & feat, const vector<KyteaString> & prefixes, KyteaModel * model, int n, int sc, int ec) {
    int w = (int)prefixes.size()/2;
    vector<KyteaChar> wind(prefixes.size());
    for(int i = w-1; i >= 0; i--)
        wind[w-i-1] = (sc-i<0?0:chars[sc-i]);
    for(int i = 0; i < w; i++)
        wind[w+i] = (ec+i>=(int)chars.length()?0:chars[ec+i]);
    unsigned ret = 0;
    for(unsigned i = 0; i < wind.size(); i++) {
        if(wind[i] == 0) continue; 
        KyteaString str = prefixes[i];
        for(int k = 0; k < n && i+k < wind.size() && wind[i+k] != 0; k++) {
            str = str+wind[i+k];
            feat.push_back(model->mapFeat(str));
            ret++;
        }
    }

    return ret;
}

class PETriplet {
public:
    vector< vector<unsigned> > first;
    vector<int> second;
    KyteaModel * third;
};
typedef std::tr1::unordered_map<KyteaString, PETriplet, KyteaStringHash> PEHash;
void Kytea::trainPE() {
    if(config_->getDebug() > 0)
        cerr << "Creating pronunciation estimation features ";
    if(dict_ == 0)
        return;
    // prepare prefixes
    vector<KyteaString> charPrefixes, typePrefixes;
    prepareNgramPrefixes(charPrefixes,typePrefixes);
    // find words that need to be modeled
    vector<DictionaryEntry*> & entries = dict_->getEntries();
    PEHash feats;
    for(unsigned i = 0; i < entries.size(); i++) {
        if(entries[i]->prons.size() > 1) {
            vector< vector<unsigned> > xs;
            vector<int> ys;
            feats[entries[i]->word].first = xs;
            feats[entries[i]->word].second = ys;
            if(entries[i]->pronMod)
                delete entries[i]->pronMod;
            entries[i]->pronMod = new KyteaModel();
            feats[entries[i]->word].third = entries[i]->pronMod;
        }
    }
    // build features
    for(Sentences::const_iterator it = sentences_.begin(); it != sentences_.end(); it++) {
        int startPos = 0, finPos=0;
        KyteaString charStr = (*it)->chars;
        KyteaString typeStr = util_->mapString(util_->getTypeString(charStr));
        for(unsigned j = 0; j < (*it)->words.size(); j++) {
            startPos = finPos;
            KyteaWord & word = (*it)->words[j];
            finPos = startPos+word.surf.length();
            if(word.peConf <= config_->getConfidence())
                continue;
            PEHash::iterator peit = feats.find(word.surf);
            if(peit != feats.end()) {
                unsigned myPron = dict_->getPronunciationID(word.surf,word.pron);
                if(myPron != 0) {
                    vector<unsigned> feat;
                    peCharFeatures(charStr, feat, charPrefixes, peit->second.third, config_->getCharN(), startPos-1, finPos);
                    peCharFeatures(typeStr, feat, typePrefixes, peit->second.third, config_->getTypeN(), startPos-1, finPos);
                    peit->second.first.push_back(feat);
                    peit->second.second.push_back(myPron);
                }
            }
        }
    }
    if(config_->getDebug() > 0)
        cerr << "done!" << endl << "Training PE classifers ";
    // calculate classifiers
    for(PEHash::iterator peit = feats.begin(); peit != feats.end(); peit++) {
        vector< vector<unsigned> > & xs = peit->second.first;
        vector<int> & ys = peit->second.second;
        peit->second.third->trainModel(xs,ys,config_->hasBias(),config_->getSolverType(),config_->getEpsilon());
    }
    if(config_->getDebug() > 0)
        cerr << "done!" << endl;
}


void Kytea::writeModels() {

    if(config_->getDebug() > 0)    
        cerr << "Printing model to " << config_->getModelFile();
    
    ModelIO * modout = ModelIO::createIO(config_->getModelFile().c_str(),config_->getModelFormat(), true, *config_);

    modout->writeConfig(*config_);
    modout->writeModel(wsModel_);
    modout->writeDictionary(dict_);

    delete modout;

    if(config_->getDebug() > 0)    
        cerr << " done!" << endl;


}

void Kytea::readModels() {
    
    if(config_->getDebug() > 0)    
        cerr << "Reading model from " << config_->getModelFile();
    
    ModelIO * modin = ModelIO::createIO(config_->getModelFile().c_str(),ModelIO::FORMAT_UNKNOWN, false, *config_);
    util_ = config_->getStringUtil();

    modin->readConfig(*config_);
    wsModel_ = modin->readModel();
    dict_ = modin->readDictionary();

    delete modin;

    if(config_->getDebug() > 0)    
        cerr << " done!" << endl;
}

void Kytea::analyzeInput() {

    CorpusIO *in, *out;
    iostream *inStr = 0, *outStr = 0;
    const vector<string> & args = config_->getArguments();
    if(args.size() > 0) {
        in  = CorpusIO::createIO(args[0].c_str(),config_->getInputFormat(), false, util_);
    } else {
        inStr = new iostream(cin.rdbuf());
        in  = CorpusIO::createIO(*inStr, config_->getInputFormat(), false, util_);
    }
    if(args.size() > 1) {
        out  = CorpusIO::createIO(args[1].c_str(),config_->getOutputFormat(), true, util_);
    } else {
        outStr = new iostream(cout.rdbuf());
        out = CorpusIO::createIO(*outStr, config_->getOutputFormat(), true, util_);
    }

    // prepare prefixes
    vector<unsigned> dictFeats;
    vector<KyteaString> charPrefixes, typePrefixes;
    if(config_->getDoWS())
        prepareDictPrefixes(dictFeats);
    prepareNgramPrefixes(charPrefixes,typePrefixes);

    KyteaSentence* next;
    while((next = in->readSentence()) != 0) {
        if(config_->getDoWS())
            calculateWS(next, dictFeats, charPrefixes, typePrefixes);
        if(config_->getDoPE())
            calculatePE(next, charPrefixes, typePrefixes);
        out->writeSentence(next);
        delete next;
    }

    delete in;
    delete out;
    if(inStr) delete inStr;
    if(outStr) delete outStr;

}

void Kytea::calculateWS(KyteaSentence * sent, const vector<unsigned> & dictFeats, const vector<KyteaString> & charPrefixes, const vector<KyteaString> & typePrefixes) {
    
    // get the features for the sentence
    SentenceFeatures feats(sent->wsConfs.size());
    wsDictionaryFeatures(sent->chars, feats, dictFeats);
    wsNgramFeatures(sent->chars, feats, charPrefixes, config_->getCharN());
    wsNgramFeatures(util_->mapString(util_->getTypeString(sent->chars)), feats, typePrefixes, config_->getTypeN());

    for(unsigned i = 0; i < sent->wsConfs.size(); i++) {
        if(abs(sent->wsConfs[i]) <= config_->getConfidence()) {
            pair<int,double> answer = wsModel_->runClassifier(feats[i]);
            sent->wsConfs[i] = answer.first * abs(answer.second);
        }
    }

    sent->refreshWS(config_->getConfidence());

}

// TODO: make this actually estimate the correct value
pair<KyteaString,double> Kytea::calculateUnknownPE(const KyteaString & str) {
    pair<KyteaString,double> ret(util_->mapString("UNK"),0);
    return ret;
}
void Kytea::calculatePE(KyteaSentence * sent, const vector<KyteaString> & charPrefixes, const vector<KyteaString> & typePrefixes) {
    int startPos = 0, finPos=0;
    KyteaString charStr = sent->chars;
    KyteaString typeStr = util_->mapString(util_->getTypeString(charStr));
    for(unsigned i = 0; i < sent->words.size(); i++) {
        KyteaWord & word = sent->words[i];
        startPos = finPos;
        finPos = startPos+word.surf.length();
        const DictionaryEntry* ent = dict_->findEntry(word.surf);
        if(ent == 0 || ent->prons.size() == 0) {
            pair<KyteaString, double> uk = calculateUnknownPE(word.surf);
            word.pron = uk.first;
            word.peConf = uk.second;
        }
        else if(ent->pronMod == 0) {
            word.pron = ent->prons[0];
            word.peConf = 100; 
        }
        else {
            vector<unsigned> feat;
            peCharFeatures(charStr, feat, charPrefixes, ent->pronMod, config_->getCharN(), startPos-1, finPos);
            peCharFeatures(typeStr, feat, typePrefixes, ent->pronMod, config_->getTypeN(), startPos-1, finPos);
            pair<int,double> answer = ent->pronMod->runClassifier(feat);
            word.pron = ent->prons[answer.first-1];
            word.peConf = answer.second;
        }
    }
}
